# -*- coding: utf-8 -*-

"""'ligands' subcommand """

# ----------------------- Imports --------------------------------
import os
from pathlib import Path
import datetime
import subprocess
import csv
import sys
import shutil

# My modules
import config
from modules.utils import verbose
from modules.utils import software
from modules.utils import molecules

#--------------------------Functions------------------------------
# Setup Phenix
if os.name == 'posix':
    env_phenix = software.shell_source(Path(config.PHENIX_PATH).joinpath("phenix_env.sh"))  # sourcer Phenix sous Linux
    shell_choice = False
elif os.name == 'nt':
    env_phenix = software.setup_bat(Path(config.PHENIX_PATH).joinpath("phenix_env.bat"))  # sourcer Phenix sous Windows
    shell_choice = True


def ligands_run(data_dir, lig_csv_file):
    """
    'ligands' subcommand
    :param data_dir: data directory
    :param lig_csv_file: ligands csv file
    :return: log message
    """

    date = datetime.datetime.now().strftime("%Y-%m-%d_%H%M%S")

    """Folder where we work"""
    work_dir = Path.cwd()
    """The existence of following path (created with 'prep' subcommand) has been checked in easypipe"""
    phenix_dir = work_dir.joinpath('easYPipe')

    """Creation of lig directory (should already exists)"""
    lig_dir = phenix_dir.joinpath('1c_ligands')
    lig_dir.mkdir(exist_ok=True)

    """Create a log file with ligands generation information"""
    lig_dir.joinpath('logs').mkdir(exist_ok=True)
    nom_lig_log = "ligands_easypipe_" + date + ".log"
    lig_log = verbose.log_start(lig_dir, 'logs', nom_lig_log, verbose.easypipe_logo)

    """Read information in the csv, and list ligand with smiles to be treated"""
    data_no_cif = []  # jeux sans cif après traitement
    data_cif_deja = []  # jeux avec déjà un cif
    data_no_info = []   # jeux sans info pour générer le cif
    data_cif_ok = []  # jeux avec cif bien généré
    data_cif_fail = []     # jeux avec échec de génération du cif
    ligand_smiles_pbm = []  # ligand avec plusieurs smiles associés
    ligand_with_error = []  # tuple (ligand, smiles) with error during generation of cif file
    verbose.print_and_write(
        verbose.cadre_text(config.motif, config.nb_motifs, "Reading information in the ligands csv file: "
                           + lig_csv_file.name), lig_log)
    # lister les ligands + smiles, enlever les doubles
    verbose.print_and_write("Ligand cif will be generated for: " , lig_log)
    ligands_smiles = {}    # dictionnaire {nom du ligand : smiles}
    with open(lig_csv_file, newline='') as csvfile:
        lig_csv_reader = csv.DictReader(csvfile)
        for row in lig_csv_reader:
            if row['ligand name'] and row['ligand smiles']:  # gérer si pas de smi dans le csv
                verbose.print_and_write(row['ligand name'] + ": " + row['ligand smiles'], lig_log)
                # vérifier si ligand déjà dans dict
                if row['ligand name'] in ligands_smiles:
                    # vérifier que le smiles est identique ou pas
                    if ligands_smiles[row['ligand name']] != row['ligand smiles']:
                        verbose.print_and_write("\tThis ligand name already exists with another smiles: "
                                                + ligands_smiles[row['ligand name']]
                                                + "\n\tit won't be generated with this smiles ...", lig_log)
                        ligand_smiles_pbm.append(row['ligand name'])
                else:
                    ligands_smiles[row['ligand name']] = row['ligand smiles']
    #print(ligands_smiles)       # TEST
    if not ligands_smiles:
        verbose.print_and_write("There is no cif to generate ...", lig_log)
        sys.exit(1)

    """Generation of pdb and cif files in a folder '1c_ligands/ligands': create smiles file and converts it with elBOW to pdb and cif"""
    # générer les pdb et cif dans un dossier 1c_ligands/ligands avec des sous-dossiers portant le nom du ligand
    verbose.print_and_write(verbose.cadre_text(config.motif, config.nb_motifs,
                                               "pdb and cif generation in '1c_ligands/ligands' folder"), lig_log)

    # création du dossier 1c_ligands/ligands
    lig_dir_calc = lig_dir.joinpath('ligands')
    lig_dir_calc.mkdir(exist_ok=True)
    for lig, smiles in ligands_smiles.items():
        verbose.print_and_write("\nLigand name: " + lig, lig_log)
        # vérifier si le sous-dossier du ligand existe et s'il y a bien un cif dedans
        ligname_dir = lig_dir_calc.joinpath(lig)
        if ligname_dir.is_dir() and list(ligname_dir.glob('*.cif')):
            verbose.print_and_write("cif file already generated for this ligand ...", lig_log)
        else:
            verbose.print_and_write("SMILES: " + smiles, lig_log)
            ligname_dir.mkdir(exist_ok=True)
            # générer le pdb et cif dans ce sous-dossier
            smi_str = smiles
            # enlever les sels
            if len(smi_str.split('.')) > 1:  # i.e. il y a des sels
                # enlever les sels
                # couper là où '.' et garder le plus gros
                smi_str_no_salt = ""
                for part in smi_str.split('.'):
                    if len(part) > len(smi_str_no_salt):
                        smi_str_no_salt = part
                verbose.print_and_write("Salt stripped, new smiles: " + smi_str_no_salt, lig_log)
            else:
                smi_str_no_salt = smi_str
            # canoniser
            smi_str_no_salt_cano, return_code, err = molecules.smi_cano(smi_str_no_salt)
            if return_code != 0:
                verbose.print_and_write("Canonization failed ...\n" + smi_str_no_salt_cano + err, lig_log)
            else:
                verbose.print_and_write("Canonized, new smiles: " + smi_str_no_salt_cano, lig_log)
            # écriture du fichier smiles valide
            smi_file_OK = ligname_dir.joinpath(lig + '.smi')
            verbose.print_and_write("Creation of ligand smiles file " + smi_file_OK.name, lig_log)
            with open(smi_file_OK, 'w') as smif:
                smif.write((smi_str_no_salt_cano if return_code == 0 else smi_str_no_salt).strip())
            # elBOW sur fichier smiles après avoir enlevé l'ID, les sels et canonisé
            os.chdir(ligname_dir)  # lancer eLBOW dans le dossier du ligand
            verbose.print_and_write("Generation of ligand cif and pdb from smiles with eLBOW.", lig_log)
            elbow = subprocess.Popen(
                ['phenix.elbow', '--smiles=' + str(smi_file_OK), '--opt', '--output=' + lig],
                stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=shell_choice, env=env_phenix,
                encoding='utf-8')
            out, err = elbow.communicate()
            if err:
                verbose.print_and_write(err, lig_log)
                ligand_with_error.append((lig, smiles))



    """Copy of pdb and cif files in each dataset working folder"""
    verbose.print_and_write(verbose.cadre_text(config.motif, config.nb_motifs,
                                               "Copy of ligand files in dataset working folder"), lig_log)
    # copie des pdb et cif dans chaque jeu
    with open(lig_csv_file, newline='') as csvfile:
        lig_csv_reader = csv.DictReader(csvfile)
        for row in lig_csv_reader:
            verbose.print_and_write("Dataset: " + row['dataset'], lig_log)
            # dossier ligand_dir déjà créé à l'étape 'prep'
            ligand_dir = phenix_dir.joinpath('0_processed_datasets', row['dataset'], 'ligand')
            # traiter chaque ligne, i.e. pour chaque dataset, traiter le ligand correspondant
            # créer le dossier 'ligand' dans le jeu
            if row['ligand name'] and row['ligand smiles']:  # gérer si pas de smi dans le csv
                cif_in_lig = [f.name for f in ligand_dir.iterdir() if f.suffix.lower() == '.cif']
                if not cif_in_lig:  # ne copie pas si déjà un cif dans le dossier ligand
                    verbose.print_and_write("Ligand name: " + row['ligand name'], lig_log)
                    verbose.print_and_write("SMILES: " + row['ligand smiles'], lig_log)
                    # chercher le dossier où a été généré le cif
                    ligname_dir = lig_dir_calc.joinpath(row['ligand name'])
                    # copie des fichiers smi, cif et pdb dans ligand_dir
                    if ligname_dir.exists():
                        files_to_copy = [f.name for f in ligname_dir.iterdir() if f.suffix.lower() in ['.cif', '.smi', '.pdb']]
                        #print(files_to_copy)        # TEST
                        verbose.print_and_write("Files copied: ", lig_log)
                        for f in files_to_copy:
                            verbose.print_and_write(f, lig_log)
                            shutil.copy2(ligname_dir.joinpath(f), ligand_dir.joinpath(f))
                    # tester ci bien cif copié
                    cif_in_lig = [f.name for f in ligand_dir.iterdir() if f.suffix.lower() == '.cif']
                    if not cif_in_lig:
                        data_cif_fail.append(row['dataset'])
                    else:
                        data_cif_ok.append(row['dataset'])
                else:
                    verbose.print_and_write("Already a cif file in 'ligand' directory: " + ", ".join(cif_in_lig),
                                            lig_log)
                    data_cif_deja.append(row['dataset'])
                    if len(cif_in_lig) > 1:  # cas où plusieurs cif
                        verbose.print_and_write(
                            "More than one cif file, please make sure there is only one cif file in 'ligand' directory, "
                            "else only the first one will be considered ...", lig_log)


            else:
                data_no_info.append(row['dataset'])
                verbose.print_and_write(
                    "No smiles for this dataset" if not row['ligand smiles'] else "No ligand name "
                                                                                  "for this dataset", lig_log)

            cif_in_lig = [f.name for f in ligand_dir.iterdir() if f.suffix.lower() == '.cif']
            if not cif_in_lig:
                data_no_cif.append(row['dataset'])
            verbose.print_and_write(verbose.chaine_xfois(config.sub_motif, config.nb_motifs),
                                    lig_log)

    """Summary"""
    verbose.print_and_write(verbose.chaine_xfois(config.motif, config.nb_motifs), lig_log)
    verbose.print_and_write(verbose.easypipe_logo, lig_log)
    verbose.print_and_write(verbose.cadre_text(config.motif, config.nb_motifs, "SUMMARY"), lig_log)
    # lister les jeux avec cif générés
    if data_cif_ok:
        verbose.print_and_write("Ligand cif and pdb have been generated from smiles with eLBOW "
                                "and are located in each dataset 'ligand' folder.", lig_log)
        verbose.print_and_write("\nDatasets with generated cif:", lig_log)
        for j in data_cif_ok:
            verbose.print_and_write(j, lig_log)
        if data_cif_fail:
            verbose.print_and_write("\nDatasets for which generation of cif has failed:", lig_log)
            for j in data_cif_fail:
                verbose.print_and_write(j, lig_log)
    else:
        verbose.print_and_write("No cif generated from smiles ...", lig_log)

    # lister les jeux qui avaient déjà un cif
    if data_cif_deja:
        verbose.print_and_write("\nDatasets already with a cif in 'ligand' folder:", lig_log)
        for j in data_cif_deja:
            verbose.print_and_write(j, lig_log)

    # lister les ligands avec problème lors de la génération du cif
    if ligand_with_error:
        verbose.print_and_write("\nBeware !! Ligand(s) with problem in cif generation:", lig_log)
        for lig, smi in ligand_with_error:
            verbose.print_and_write(lig + ": " + smi, lig_log)
        verbose.print_and_write("!! Please correct the smiles in csv before running again.", lig_log)

    # lister les ligands avec même nom pour 2 smiles différents
    if ligand_smiles_pbm:
        verbose.print_and_write("\nBeware !! Ligand(s) with 2 different smiles for the same name:", lig_log)
        for lig in ligand_smiles_pbm:
            verbose.print_and_write(lig, lig_log)
        verbose.print_and_write("!! Please delete all generated files for this ligand name "
                                "in '1c_ligands/ligands' and '0_processed_datasets' + dataset + ligand name,"
                                " and correct smiles in csv before running again.", lig_log)


    # lister les jeux sans info pour générer le cif
    if data_no_info:
        verbose.print_and_write("\nDatasets without smiles information for cif generation:", lig_log)
        for j in data_no_info:
            verbose.print_and_write(j, lig_log)

    # lister les jeux sans cif
    if data_no_cif:
        verbose.print_and_write(verbose.chaine_xfois(config.sub_motif, config.nb_motifs), lig_log)
        verbose.print_and_write("\nDatasets without cif file for ligand search:", lig_log)
        for j in data_no_cif:
            verbose.print_and_write(j, lig_log)
        verbose.print_and_write(
            "\nYou can add manually a cif file in 'ligand' directory of corresponding dataset (in '0_processed_datasets' folder) "
            "if you don't want to generate it from smiles or if generation with elBOW has failed ...\n", lig_log)
    os.chdir(work_dir)
    return lig_log